%function [ U ] = pdecol( varargin )
% PDECOL - Solve a PDE with algorithm 540 from the (collocation method).
% 
% Written by:
% -- 
% John L. Weatherwax                2005-04-27
% 
% email: wax@alum.mit.edu
% 
% Please send comments and especially bug reports to the
% above email address.
% 
%-----

% Any ideas on how to get the number of pdes's to solve without having to pass this number in?
npde                        = 2;   
%npde                        = 1;   

% Default pdecol input arguments: 
%  
t0                          = 0.0;      % The initial time
xL                          = 0;        % The left end of the computational domain
xR                          = 1;        % The right end of the computational domain
nint                        = 30;       % The number of intervals to divide the domain into
dt                          = 1e-7;     % The initial guess at the timestep (WWX: Check)
eps                         = 1e-4;     % The desired relative accuracy (WWX: Check)
kord                        = 4;        % The order of the B-Spline colocation approximation (from 3-20) (WWX: Check)
ncc                         = 2;        % The number of B-Spline enforced continuity conditions (WWX: Check)
%                                         The location of the collocation points (not required to be uniformly spaced):
xbkpt                       = linspace( xL, xR, nint+1 ); 
xGrid                       = xbkpt;    % The grid locations of the desired solution (no special requirements)

% Output times the solution is desired:
%tGrid                       = [ t0 1d-3 ];
tGrid                       = [ 0 10.^[ -3:1 ] ];
%tGrid                       = linspace( t0, t0+2d-3, 10 );
%tGrid                       = linspace( t0, 0.1, 2 ); 

%
% MF = 10*METH + MITER
% 
% METH=1 (ADAMS method (a generilization of crank-nicolson)
%     =2 (BACKWARDS DIFFERENTIATION FORMULAS (BDF) OR STIFF METHOD OF GEAR)
% MITER=1 (CHORD METHOD WITH ANALYTIC JACOBIANS, USER SUPPLIES SUBROUTINE DERIVF
%      =2 CHORD METHOD CALCULATED WITH FINITE DIFFERENCES
%
%mf                          = 12;
%mf                          = 22; 

%mf                          = 11; 
mf                          = 21; 

% Number of desired "u" (dependent variable) derivatives to compute:
nderv = 0; 

% Override default arguments with "varargin" inputs:
%
i=1; 
while i <= nargin
  arg = varargin{i}
  switch lower(arg)
   case 'xl'
    xL = varargin{i+1};
   case 'xr'
    xR = varargin{i+1};
   case 'epsilon' 
    eps = varargin{i+1};
   case 'kord'
    kord = varargin{i+1};
   case 'ncc'
    ncc = varargin{i+1};
   case 'xbkpt'
    xbkpt = varargin{i+1};
   case 'tGrid'
    tGrid = varargin{i+1};
   case 'mf'
    mf = varargin{i+1};
   otherwise
    error( [ 'Unknown input', arg ] ); 
  end
  i=i+2;
end

% Input argument error checking:
%
if( xL >= xR ) 
  error( 'xL >= xR' ); 
end
if( eps <= 0 )
  error( 'eps <= 0' ); 
end
if( kord < 3 | kord > 20 )
  error( 'kord < 3 | kord > 20' ); 
end
if( ncc <= 1 | ncc > kord )
  error( 'ncc <= 1 | ncc > kord' ); 
end
if( nint < 1 )
  error( 'nint < 1' ); 
end
if( length(tGrid) < 2 ) 
  error( 'length(tGrid) < 2' ); 
end
if( nderv >= kord ) 
  error( 'nderv >= kord' ); 
end

% Calculate the required auxiliary data arrays sizes:
%
nbkpt       = nint+1;
ncpts       = kord*nint - ncc*(nint-1);
iquad       = 1;                          % assume the larger value (v.s. 0)
ml          = npde*(kord+iquad-1) - 1;
maxder      = 5;

nwork       = kord + 4*npde + 9*npde^2 + ncpts*(3*kord + 2) + ...
                                    npde*ncpts*(3*ml + maxder + 7);
niwork      = ncpts*(npde+1);
nsctch      = kord*(nderv+1);

% Call the octave *.oct file that calls the FORTRAN pdecol.f:
%
disp( 'entering pdecolg ...' );
% Without Jacobians...
% $$$ [U] = pdecolg(tGrid,dt,xbkpt,eps,kord,ncc,mf,npde,nwork,niwork,nsctch,...
% $$$               "F","DBDU_BNDRY","DBDUX_BNDRY","DZDT_BNDRY","UINIT");
[U] = pdecolg(tGrid,dt,xbkpt,eps,kord,ncc,mf,npde,nwork,niwork,nsctch,...
              "F","DBDU_BNDRY","DBDUX_BNDRY","DZDT_BNDRY","UINIT",...
              "DFDU_DERIVF","DFDUX_DERIVF","DFDUXX_DERIVF")

return; 

if( 0 ) 
% Plot the solution U:
h1=figure; plot( xbkpt, U(1,:), '-x' ); grid; title( [ 'U(1) '] ); 
pause;
h2=figure; plot( xbkpt, U(2,:), '-x' ); grid; title( [ 'U(2) '] ); 
pause;
return; 
end

close all; 
ntouts = length(tGrid)-1; 
h=figure; 
for nti=1:ntouts,
  figure(h); 
  subplot( 1,2,1 ); %axis( [ 0 1 0.4 1.8 ] ); 
  plot( xbkpt, U(1,:,nti) ); grid; %title( [ 'U(1) at t(',num2str(nti),')' ] ); 
  %pause; 
  subplot( 1,2,2 ); %axis( [ 0 1 2.7 3.7 ] ); 
  plot( xbkpt, U(2,:,nti) ); grid; %title( [ 'U(2) at t(',num2str(nti),')' ] ); 
  pause;
end

% Collect all inputs together into a structure:
%
opts.npde           = npde;

opts.t0             = t0;
opts.xL             = xL;
opts.xR             = xR; 
opts.nint           = nint; 
opts.dt             = dt;
opts.eps            = eps;
opts.kord           = kord;
opts.ncc            = ncc;
opts.xbkpt          = xbkpt;
opts.tGrid          = tGrid;
opts.mf             = mf;

opts.nderv          = nderv;
