/*

Copyright (C) 2005, 2006 John L. Weatherwax

This file is part of Octave.

Octave is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

Octave is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Octave; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   
Written by:
-- 
John L. Weatherwax                2005-05-15
   
email: wax@alum.mit.edu
   
Please send comments and especially bug reports to the
above email address.
   
*/

#include <octave/oct.h>
#include <octave/parse.h>
#include <octave/f77-fcn.h>

// These are the declarations of the routines that the FORTRAN code will call:
//
typedef F77_RET_T (*pc_f_ptr) (const double&, const double&, double*, double*, double*, double*, const int& );
typedef F77_RET_T (*pc_bndry_ptr) (const double&, const double&, double*, double*, double*, double*, double*, const int& );
typedef F77_RET_T (*pc_uinit_ptr) (const double&, double*, const int& );
typedef F77_RET_T (*pc_derivf_ptr) (const double&, const double&, double*, double*, double*, double*, double*, double*, const int& );

// These are the declarations of the FORTRAN subroutine that the octave C++ gateway routine
// will call:
//
extern "C" {
  int F77_FUNC(looppdecol,LOOPPDECOL) ( double*, int&, double&, double*, int&, \
                                        double&, int&, int&, int&, int&, \
                                        double*, int*, double*, double* );
  int F77_FUNC(pdecol,PDECOL) ( double&, double&, double&, double*, double&, \
				int&, int&, int&, int&, int&, int&, \
				double*, int*, \
                                pc_f_ptr, pc_bndry_ptr, pc_uinit_ptr, pc_derivf_ptr ); 
  int F77_FUNC(values,VALUES) ( double*, double*, double*, int&, int&, int&, int&, double* );
}

// Global static octave function pointers for the user defined functions
// required by pdecol:
//
static octave_function *pdecol_f_fcn = 0;
static octave_function *pdecol_dbdu_fcn = 0;
static octave_function *pdecol_dbdux_fcn = 0; 
static octave_function *pdecol_dzdt_fcn = 0; 
static octave_function *pdecol_uinit_fcn = 0;
static octave_function *pdecol_dfdu_fcn = 0; 
static octave_function *pdecol_dfdux_fcn = 0;
static octave_function *pdecol_dfduxx_fcn = 0;

// Here are two things (one after each other):
// 
// 1) C++ octave API calls of each user function:
// --These enable calls of the user functions from within the C++
//   environment ... mainly for testing
// 2) C++ octave class <-> FORTRAN conversion subroutine/function
// --These are the functional prototypes called by the FORTRAN
// 


//
// Code for "EXTERNAL F"
// 
ColumnVector
pdecol_f_function (double t, double x, const ColumnVector& u, 
                   const ColumnVector & ux, const ColumnVector& uxx)
{
  ColumnVector retval;

  octave_value_list args;
  args(0) = t; args(1) = x; args(2) = u; args(3) = ux; args(4) = uxx; 

  if (pdecol_f_fcn)
    {
      octave_value_list tmp = pdecol_f_fcn->do_multi_index_op (1, args);

      if (error_state)
	{
	  gripe_user_supplied_eval ("pdecol_f");
	  return retval;
	}

      if (tmp.length () > 0 && tmp(0).is_defined ())
	{
	  retval = ColumnVector (tmp(0).vector_value ());

	  if (error_state || retval.length () == 0)
	    gripe_user_supplied_eval ("pdecol_f");
	}
      else
	gripe_user_supplied_eval ("pdecol_f");
    }

  return retval;
}

// This is the subroutine that will be compiled and called by the FORTRAN driver
// routine pdecol.f.
static F77_RET_T
pdecol_fortran_f_func(const double& T, const double& X, 
		      double* U, double* UX, double* UXX, double* FVAL, const int& NPDE)
{
  octave_value_list args;
  args(0) = T; args(1) = X; 

  ColumnVector Uv(NPDE),UXv(NPDE),UXXv(NPDE);
  for( int i=0; i < NPDE; i++ ){
    Uv(i)=U[i]; UXv(i)=UX[i]; UXXv(i)=UXX[i];
  }
  args(2) = Uv; args(3) = UXv; args(4) = UXXv; 
  int nargout = 1; 

  //octave_stdout << "In pdecol_fortran_f_func before feval" << "\n";
  octave_value_list retval = feval(pdecol_f_fcn, args, nargout);
  if (! error_state && retval.length () == 1){
      Matrix FVALv = retval(0).matrix_value(); 
      if( !error_state ){
	for (int i = 0; i < NPDE; i++){
	  //octave_stdout << "FVALv(" << i << ") = " << FVALv(i) << "\n"; 
	  FVAL[i] = FVALv(i);
	}
      }else{
	error("pfff: failure evaluating user-supplied function");
      }
  }
  else
    error ("pfff: failure evaluating user-supplied function");

  F77_RETURN(0);
}


//
// Code for "EXTERNAL BNDRY"
//
Matrix
pdecol_dbdu_function (double t, double x, const ColumnVector& u, const ColumnVector & ux)
{
  Matrix retval;

  octave_value_list args;
  args(0) = t; args(1) = x; args(2) = u; args(3) = ux;

  if (pdecol_dbdu_fcn)
    {
      octave_value_list tmp = pdecol_dbdu_fcn->do_multi_index_op (1, args);

      if (error_state)
	{
	  gripe_user_supplied_eval ("pdecol_dbdu");
	  return retval;
	}

      if (tmp.length () > 0 && tmp(0).is_defined ())
	{
	  retval = Matrix (tmp(0).matrix_value ());

	  if (error_state || retval.length () == 0)
	    gripe_user_supplied_eval ("pdecol_dbdu");
	}
      else
	gripe_user_supplied_eval ("pdecol_dbdu");
    }

  return retval;
}

Matrix
pdecol_dbdux_function (double t, double x, const ColumnVector& u, const ColumnVector & ux)
{
  Matrix retval;

  octave_value_list args;
  args(0) = t; args(1) = x; args(2) = u; args(3) = ux;

  if (pdecol_dbdux_fcn)
    {
      octave_value_list tmp = pdecol_dbdux_fcn->do_multi_index_op (1, args);

      if (error_state)
	{
	  gripe_user_supplied_eval ("pdecol_dbdux");
	  return retval;
	}

      if (tmp.length () > 0 && tmp(0).is_defined ())
	{
	  retval = Matrix (tmp(0).matrix_value ());

	  if (error_state || retval.length () == 0)
	    gripe_user_supplied_eval ("pdecol_dbdux");
	}
      else
	gripe_user_supplied_eval ("pdecol_dbdux");
    }

  return retval;
}

ColumnVector
pdecol_dzdt_function (const double& t, const double& x, const ColumnVector& u, const ColumnVector & ux)
{
  ColumnVector retval;

  octave_value_list args;
  args(0) = t; args(1) = x; args(2) = u; args(3) = ux;

  if (pdecol_dzdt_fcn)
    {
      octave_value_list tmp = pdecol_dzdt_fcn->do_multi_index_op (1, args);

      if (error_state)
	{
	  gripe_user_supplied_eval ("pdecol_dzdt");
	  return retval;
	}

      if (tmp.length () > 0 && tmp(0).is_defined ())
	{
	  retval = ColumnVector (tmp(0).vector_value ());

	  if (error_state || retval.length () == 0)
	    gripe_user_supplied_eval ("pdecol_dzdt");
	}
      else
	gripe_user_supplied_eval ("pdecol_dzdt");
    }

  return retval;
}

// This is the subroutine that will be compiled and called by the FORTRAN driver
// routine pdecol.f.
static F77_RET_T
pdecol_fortran_bndry_func(const double& T, const double& X, 
			  double* U, double* UX, 
			  double* DBDU, double* DBDUX, double* DZDT, const int& NPDE){
  octave_value_list args;

  // Prepare for a calls to pdecol_dbdu_fcn, pdecol_dbdux_fcn, and pdecol_dzdt_fcn: 
  args(0) = T; args(1) = X; 
  ColumnVector Uv(NPDE),UXv(NPDE);
  for( int i=0; i < NPDE; i++ ){
    Uv(i)=U[i]; UXv(i)=UX[i];
  }
  args(2) = Uv; args(3) = UXv; 
  int nargout = 1; 

  octave_value_list retval; 

  //octave_stdout << "In pdecol_fortran_bndry_func before feval to DBDU_BNDRY" << "\n";
  retval = feval(pdecol_dbdu_fcn, args, nargout);
  if( error_state || retval.length() != 1 ){
    error("failure evaluating user-supplied function DBDU_BNDRY");
  }
  Matrix DBDUm = retval(0).matrix_value(); 
  for (int i = 0; i < NPDE*NPDE; i++){
    //octave_stdout << "DBDUm(" << i << ") = " << DBDUm(i) << "\n"; 
    DBDU[i] = DBDUm(i);
  }
  
  //octave_stdout << "In pdecol_fortran_bndry_func before feval to DBDUX_BNDRY" << "\n";
  retval = feval(pdecol_dbdux_fcn, args, nargout);
  if( error_state || retval.length() != 1 ){
    error("failure evaluating user-supplied function DBDUX_BNDRY");
  }
  Matrix DBDUXm = retval(0).matrix_value(); 
  for (int i = 0; i < NPDE*NPDE; i++){
    //octave_stdout << "DBDUXm(" << i << ") = " << DBDUXm(i) << "\n"; 
    DBDUX[i] = DBDUXm(i);
  }
  
  //octave_stdout << "In pdecol_fortran_bndry_func before feval to DZDT_BNDRY" << "\n";
  retval = feval(pdecol_dzdt_fcn, args, nargout);
  if( error_state || retval.length() != 1 ){
    error("failure evaluating user-supplied function DZDT_BNDRY");
  }
  Matrix DZDTm = retval(0).matrix_value(); 
  for (int i = 0; i < NPDE; i++){
    //octave_stdout << "DZDTm(" << i << ") = " << DZDTm(i) << "\n"; 
    DZDT[i] = DZDTm(i);
  }

  F77_RETURN(0);
}


//
// Code for "EXTERNAL UINIT"
// 
ColumnVector
pdecol_uinit_function (double x)
{
  ColumnVector retval;

  octave_value_list args;
  args(0) = x;

  if (pdecol_uinit_fcn)
    {
      octave_value_list tmp = pdecol_uinit_fcn->do_multi_index_op (1, args);

      if (error_state)
	{
	  gripe_user_supplied_eval ("pdecol_uinit");
	  return retval;
	}

      if (tmp.length () > 0 && tmp(0).is_defined ())
	{
	  retval = ColumnVector (tmp(0).vector_value ());

	  if (error_state || retval.length () == 0)
	    gripe_user_supplied_eval ("pdecol_uinit");
	}
      else
	gripe_user_supplied_eval ("pdecol_uinit");
    }

  return retval;
}

// This is the subroutine that will be compiled and called by the FORTRAN driver
// routine pdecol.f.
static F77_RET_T
pdecol_fortran_uinit_func(const double& X, double* U, const int& NPDE){
  octave_value_list args;

  // Prepare for a calls to pdecol_uinit_fcn:
  args(0) = X; 
  int nargout = 1; 

  //octave_stdout << "In pdecol_fortran_uinit_func before feval to UINIT" << "\n";
  octave_value_list retval = feval(pdecol_uinit_fcn, args, nargout);
  if( error_state || retval.length() != 1 ){
    error("failure evaluating user-supplied function DBDU_BNDRY");
  }
  Matrix Um = retval(0).matrix_value(); 
  for (int i = 0; i < NPDE; i++){
    //octave_stdout << "Um(i)=" << Um(i) << "\n";
    U[i] = Um(i);
  }

  F77_RETURN(0);
}


//
// Code for "EXTERNAL DERIVF"
//
Matrix
pdecol_dfdu_function (double t, double x, const ColumnVector& u, 
                      const ColumnVector & ux, const ColumnVector & uxx)
{
  Matrix retval;

  octave_value_list args;
  args(0) = t; args(1) = x; args(2) = u; args(3) = ux; args(4) = uxx; 

  if (pdecol_dfdu_fcn)
    {
      octave_value_list tmp = pdecol_dfdu_fcn->do_multi_index_op (1, args);

      if (error_state)
	{
	  gripe_user_supplied_eval ("pdecol_dfdu");
	  return retval;
	}

      if (tmp.length () > 0 && tmp(0).is_defined ())
	{
	  retval = Matrix (tmp(0).matrix_value ());

	  if (error_state || retval.length () == 0)
	    gripe_user_supplied_eval ("pdecol_dfdu");
	}
      else
	gripe_user_supplied_eval ("pdecol_dfdu");
    }

  return retval;
}

Matrix
pdecol_dfdux_function (double t, double x, const ColumnVector& u, 
                      const ColumnVector & ux, const ColumnVector & uxx)
{
  Matrix retval;

  octave_value_list args;
  args(0) = t; args(1) = x; args(2) = u; args(3) = ux; args(4) = uxx; 

  if (pdecol_dfdux_fcn)
    {
      octave_value_list tmp = pdecol_dfdux_fcn->do_multi_index_op (1, args);

      if (error_state)
	{
	  gripe_user_supplied_eval ("pdecol_dfdux");
	  return retval;
	}

      if (tmp.length () > 0 && tmp(0).is_defined ())
	{
	  retval = Matrix (tmp(0).matrix_value ());

	  if (error_state || retval.length () == 0)
	    gripe_user_supplied_eval ("pdecol_dfdux");
	}
      else
	gripe_user_supplied_eval ("pdecol_dfdux");
    }

  return retval;
}

Matrix
pdecol_dfduxx_function (double t, double x, const ColumnVector& u, 
                        const ColumnVector & ux, const ColumnVector & uxx)
{
  Matrix retval;

  octave_value_list args;
  args(0) = t; args(1) = x; args(2) = u; args(3) = ux; args(4) = uxx; 

  if (pdecol_dfduxx_fcn)
    {
      octave_value_list tmp = pdecol_dfduxx_fcn->do_multi_index_op (1, args);

      if (error_state)
	{
	  gripe_user_supplied_eval ("pdecol_dfduxx");
	  return retval;
	}

      if (tmp.length () > 0 && tmp(0).is_defined ())
	{
	  retval = Matrix (tmp(0).matrix_value ());

	  if (error_state || retval.length () == 0)
	    gripe_user_supplied_eval ("pdecol_dfduxx");
	}
      else
	gripe_user_supplied_eval ("pdecol_dfduxx");
    }

  return retval;
}

// This is the subroutine that will be compiled and called by the FORTRAN driver
// routine pdecol.f.
static F77_RET_T
pdecol_fortran_derivf_func(const double& T, const double& X, 
			   double* U, double* UX, double* UXX, 
			   double* DFDU, double* DFDUX, double* DFDUXX, const int& NPDE){
  octave_value_list args;

  // Prepare for a calls to pdecol_dfdu_fcn, pdecol_dfdux_fcn, and pdecol_dfduxx_fcn:
  args(0) = T; args(1) = X; 
  ColumnVector Uv(NPDE),UXv(NPDE),UXXv(NPDE);
  for( int i=0; i < NPDE; i++ ){
    Uv(i)=U[i]; UXv(i)=UX[i]; UXXv(i)=UXX[i];
  }
  args(2) = Uv; args(3) = UXv; args(4) = UXXv; 
  int nargout = 1; 

  octave_value_list retval; 

  //octave_stdout << "In pdecol_fortran_derivf_func before feval to DFDU_DERIVF" << "\n";
  retval = feval(pdecol_dfdu_fcn, args, nargout);
  if( error_state || retval.length() != 1 ){
    error("pfdf: failure evaluating user-supplied function DFDU_DERIVF");
  }
  Matrix DFDUm = retval(0).matrix_value(); 
  for (int i = 0; i < NPDE*NPDE; i++){
    //octave_stdout << "DFDUm(" << i << ") = " << DFDUm(i) << "\n"; 
    DFDU[i] = DFDUm(i);
  }
  
  //octave_stdout << "In pdecol_fortran_derivf_func before feval to DFDUX_DERIVF" << "\n";
  retval = feval(pdecol_dfdux_fcn, args, nargout);
  if( error_state || retval.length() != 1 ){
    error("pfdf: failure evaluating user-supplied function DFDUX_DERIVF");
  }
  Matrix DFDUXm = retval(0).matrix_value(); 
  for (int i = 0; i < NPDE*NPDE; i++){
    //octave_stdout << "DFDUXm(" << i << ") = " << DFDUXm(i) << "\n"; 
    DFDUX[i] = DFDUXm(i);
  }
  
  //octave_stdout << "In pdecol_fortran_derivf_func before feval to DFDUXX_DERIVF" << "\n";
  retval = feval(pdecol_dfduxx_fcn, args, nargout);
  if( error_state || retval.length() != 1 ){
    error("pfdf: failure evaluating user-supplied function DFDUXX_DERIVF");
  }
  Matrix DFDUXXm = retval(0).matrix_value(); 
  for (int i = 0; i < NPDE*NPDE; i++){
    //octave_stdout << "DFDUXXm(" << i << ") = " << DFDUXXm(i) << "\n";
    DFDUXX[i] = DFDUXXm(i);
  }

  F77_RETURN(0);
}

DEFUN_DLD(pdecolg, args, nargout, "pdecol: an algorithm for solving one dimensional PDE's of\
the form\n\
u_t = F(t,x,u,u_x,u_xx)"){

  octave_value_list retval; 

  // Input argument checking: 
  int nargin=args.length(); 
  if( nargin != 16 && nargin != 19 ){
    error("pdecolg: Incorrect number of input arguments");
    print_usage("pdecolg"); 
    return retval;
  }
  if( nargout != 1 ) {
    error( "pdecolg: Incorrect number of output arguments");
    print_usage("pdecolg"); 
    return retval;
  }

  //
  // Extract all inputs:
  //
  RowVector touts( args(0).vector_value() );
  double t0 = touts(0); // the initial time
  double *toutsP = touts.fortran_vec();
  int ntouts = args(0).length()-1;
  double DT = args(1).double_value();
  RowVector xbkpt( args(2).vector_value() );
  double *XBKPTP = xbkpt.fortran_vec();
  int npts = args(2).length();
  double EPS = args(3).double_value();
  int KORD = args(4).int_value();
  int NCC = args(5).int_value(); 
  int MF = args(6).int_value();
  int NPDE = args(7).int_value();
  // Create storage for work/iwork to pass into FORTRAN code:
  int nwork = args(8).int_value();
  OCTAVE_LOCAL_BUFFER( double, WORKP, nwork ); 
  int niwork = args(9).int_value();
  OCTAVE_LOCAL_BUFFER( int, IWORKP, niwork );
  IWORKP[0] = nwork; 
  IWORKP[1] = niwork; 
  int nsctch = args(10).int_value(); 
  OCTAVE_LOCAL_BUFFER( double, SCTCHP, nsctch ); 

  // testing the user supplied functions...debugging...
  octave_value f_arg;
  ColumnVector testV;
  Matrix testM; 

  // Extract a octave function pointer to the "F" function:
  // 
  f_arg = args(11);
  pdecol_f_fcn = extract_function
	    (f_arg, "pdecol", "__pdecol_f_fcn__",
	     "function F = __pdecol_f_fcn__ (t, x, u, ux, uxx) F = ",
	     "; endfunction");
#ifdef _TESTING_ 
  testV = pdecol_f_function(0.0,1.0,xbkpt.transpose(),xbkpt.transpose(),xbkpt.transpose());
  octave_stdout << "testV = \n" << testV << "\n"; 
#endif

  // Extract a octave function pointer to the "DBDU" function:
  // 
  f_arg = args(12);
  pdecol_dbdu_fcn = extract_function
         (f_arg, "pdecol", "__pdecol_dbdu_fcn__",
          "function DBDU = __pdecol_dbdu_fcn__ (t, x, u, ux) DBDU = ",
          "; endfunction"); 
#ifdef _TESTING_
  // See if I can call the user function "DBDU":
  testM = pdecol_dbdu_function(0.0,1.0,xbkpt.transpose(),xbkpt.transpose());
  octave_stdout << "testM = \n" << testM << "\n"; 
#endif

  // Extract a octave function pointer to the "DBDUX" function:
  // 
  f_arg = args(13);
  pdecol_dbdux_fcn = extract_function
         (f_arg, "pdecol", "__pdecol_dbdux_fcn__",
          "function DBDUX = __pdecol_dbdux_fcn__ (t, x, u, ux) DBDUX = ",
          "; endfunction"); 
#ifdef _TESTING_
  // See if I can call the user function "DBDUX":
  testM = pdecol_dbdux_function(0.0,1.0,xbkpt.transpose(),xbkpt.transpose());
  octave_stdout << "testM = \n" << testM << "\n"; 
#endif

  // Extract a octave function pointer to the "DZDT" function:
  // 
  f_arg = args(14);
  pdecol_dzdt_fcn = extract_function
         (f_arg, "pdecol", "__pdecol_dzdt_fcn__",
          "function DZDT = __pdecol_dzdt_fcn__ (t, x, u, ux) DZDT = ",
          "; endfunction"); 
#ifdef _TESTING_
  // See if I can call the user function "DZDT":
  testV = pdecol_dzdt_function(0.0,1.0,xbkpt.transpose(),xbkpt.transpose());
  octave_stdout << "testV = \n" << testV << "\n"; 
#endif

  // Extract a octave function pointer to the "UINIT" function:
  // 
  f_arg = args(15);
  pdecol_uinit_fcn = extract_function
         (f_arg, "pdecol", "__pdecol_uinit_fcn__",
          "function UINIT = __pdecol_uinit_fcn__ (x) UINIT = ",
          "; endfunction"); 
  // Call the function "UINIT" to determine the number of equations we are integrating:
  // Assume that the midpoint of the xbkpt vector is an allowable evaluation point:
  testV = pdecol_uinit_function( 0.5*(xbkpt(0) + xbkpt(xbkpt.length()-1)) );
  //#ifdef _TESTING_
  octave_stdout << "testV = \n" << testV << "\n"; 
  //#endif

  if( nargin > 16 ){ // The user provided a jacobian...
    // Extract a octave function pointer to the PDECOL DFDU function:
    // 
    f_arg = args(16);
    pdecol_dfdu_fcn = extract_function
           (f_arg, "pdecol", "__pdecol_dfdu_fcn__",
            "function DFDU = __pdecol_dfdu_fcn__ (t, x, u, ux, uxx) DFDU = ",
            "; endfunction"); 
#ifdef _TESTING_
    // See if I can call the user function "DFDU":
    testM = pdecol_dfdu_function(0.0,1.0,xbkpt.transpose(),xbkpt.transpose(),xbkpt.transpose());
    octave_stdout << "testM = \n" << testM << "\n";
#endif 

    // Extract a octave function pointer to the "DFDUX" function:
    // 
    f_arg = args(17);
    pdecol_dfdux_fcn = extract_function
           (f_arg, "pdecol", "__pdecol_dfdux_fcn__",
            "function DFDUX = __pdecol_dfdux_fcn__ (t, x, u, ux, uxx) DFDUX = ",
            "; endfunction"); 
#ifdef _TESTING_
    // See if I can call the user function "DFDUX":
    testM = pdecol_dfdux_function(0.0,1.0,xbkpt.transpose(),xbkpt.transpose(),xbkpt.transpose());
    octave_stdout << "testM = \n" << testM << "\n"; 
#endif

    // Extract a octave function pointer to the "DFDUXX" function:
    // 
    f_arg = args(18);
    pdecol_dfduxx_fcn = extract_function
           (f_arg, "pdecol", "__pdecol_dfduxx_fcn__",
            "function DFDUXX = __pdecol_dfduxx_fcn__ (t, x, u, ux, uxx) DFDUX = ",
            "; endfunction"); 
#ifdef _TESTING_
    // See if I can call the user function "DFDU":
    testM = pdecol_dfduxx_function(0.0,1.0,xbkpt.transpose(),xbkpt.transpose(),xbkpt.transpose());
    octave_stdout << "testM = \n" << testM << "\n"; 
#endif
  }

  // Check user inputs for correctness: 
  if( !(MF==11 || MF==12 || MF==21 || MF==22) ){
    error( "Incorrect value for MF.  Choices are 11, 12, 21, 22." ); 
    return retval; 
  }
  if( (MF==11 || MF==21) && (pdecol_dfdu_fcn==0 || pdecol_dfdux_fcn==0 || pdecol_dfduxx_fcn==0) ){
    error( "Functions dfdu, dfdux, and dfduxx must be supplied with this value of MF." ); 
    return retval; 
  }

  // Output Solution U is declared as an "NDArray":
  NDArray uT( dim_vector(NPDE,npts,ntouts) ); 
  Array<int> dim(3,0);

  // Call SUBROUTINE "LOOPPDECOL":
  // F77_FUNC(looppdecol,LOOPPDECOL)(toutsP,ntouts,dt,xbkptP,npts,\
  //                                 eps,kord,ncc,mf, NPDE,\
  //                                 workP,iworkP,sctchP,uMP);
  //
  // Loop, calling pdecol repeatedly:
  // 
  int NINT=npts-1; 
  int INDEX=1;
  for( int tcnt = 0; tcnt < ntouts; tcnt++ ){
    // CALL SUBROUTINE "PDECOL":
    double T0=touts(tcnt);
    double TOUT=touts(tcnt+1);
    //octave_stdout << "Before pdecol call..." << "\n";
    //octave_stdout << "T0=" << T0 << "; TOUT=" << TOUT << "\n";
    F77_FUNC(pdecol,PDECOL)(T0,TOUT,DT,XBKPTP,EPS,NINT,KORD,NCC,NPDE,MF,\
			    INDEX,WORKP,IWORKP,\
			    pdecol_fortran_f_func,pdecol_fortran_bndry_func,\
                            pdecol_fortran_uinit_func,pdecol_fortran_derivf_func);
    // CALL SUBROUTINE "VALUES":
    int derivOrder = 0; /* only output function values... */ 
    Matrix uTS( NPDE, npts );
    double *uTSP = uTS.fortran_vec(); 
    F77_FUNC(values,VALUES)(XBKPTP,uTSP,SCTCHP,NPDE,npts,npts,derivOrder,WORKP);

    // Assign the location where this page of results goes: 
    dim(2)=tcnt; 
    uT = uT.concat(uTS,dim); 
  }
  // Store to pass out:
  retval.append(uT);
  return retval;
}
